# Django
from django.apps import apps
from django.forms import modelform_factory
from django.views.generic import View
from django.http import JsonResponse
from django.template.loader import render_to_string

# Local
from ..sites import site


class RenderFieldView(View):
    http_method_names = ["get"]

    def get_form_class(self, **kwargs):
        self.field_name = kwargs.get("field")
        app_name = kwargs.get("app")
        model_name = kwargs.get("model")
        model = apps.get_model(app_name, model_name)

        if site.is_registered(model):
            modelsite = site.get_modelsite(model)
            if modelsite.form_class:
                if self.field_name in modelsite.form_class._meta.fields:
                    return modelsite.form_class
        form_class = modelform_factory(model, fields=(self.field_name,))
        return form_class

    def get(self, request, *args, **kwargs):
        form_class = self.get_form_class(**kwargs)
        form = form_class()
        context = {"field": form[self.field_name]}
        template = render_to_string("superadmin/field.html", context=context)
        res = {"template": template}
        return JsonResponse(res, status=200)



"""Mixins for formsets"""

# Django
from django.shortcuts import redirect


class Inlines:
    inlines = []

    def __init__(self, inline_classes, **kwargs):
        self.data = kwargs
        self.inlines = list(map(self.create, inline_classes))

    def create(self, inline_class):
        related_initial = None
        if "initial" in self.data and "related_initial" in self.data["initial"]:
            related_initial = self.data.get("initial").get("related_initial")
        self.data["initial"] = (
            related_initial.get(inline_class.model, []) if related_initial else []
        )
        inline = inline_class(**self.data)
        inline.extra += len(self.data["initial"])
        inline.headers = [
            field.label
            for field in inline.empty_form.visible_fields()
            if field.name in inline_class.form.Meta.fields
        ]
        return inline

    def is_valid(self):
        return all([inline.is_valid() for inline in self.inlines])

    def __iter__(self):
        for inline in self.inlines:
            yield inline


class InlinesMixin:
    """Class for add multiple inlines in form"""

    inlines = list()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        inlines = self.get_inlines()
        if isinstance(self.inlines, dict):
            for key, inline_class in self.inlines.items():
                inline = list(filter(lambda il: isinstance(il, inline_class), inlines))[
                    0
                ]
                inlines = list(filter(lambda il: not il == inline, inlines))
                context.update({key: inline})
        else:
            context.update({inline.prefix: inline for inline in inlines})
        context["inlines"] = inlines
        return context

    def get_inlines(self):
        """Method to get all formsets"""
        kwargs = self.get_form_kwargs()
        inline_classes = (
            self.inlines
            if isinstance(self.inlines, (list, tuple))
            else self.inlines.values()
        )
        inlines = Inlines(inline_classes, **kwargs)
        return inlines

    def post(self, request, *args, **kwargs):
        self.object = self.get_object() if self.action == "update" else None
        form = self.get_form()
        inlines = self.get_inlines()
        form.inlines = inlines
        if form.is_valid() and inlines.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def form_valid(self, form):
        self.object = form.save()
        for inline in form.inlines:
            inline.instance = self.object
            inline.save()
        return redirect(self.get_success_url())

    def form_invalid(self, form):
        for inline in form.inlines:
            for error in inline.errors:
                form.errors.update(error)
        return super().form_invalid(form)



""" Hydra model admin """

"""

# Django
from django.contrib import admin
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

# Models
from hydra.models import Action, Menu

# Forms
from hydra.forms import BaseActionForm, BaseMenuForm, BasePermissionForm

@admin.register(Action)
class ActionAdmin(admin.ModelAdmin):
    form = BaseActionForm
    list_display = ("__str__", "app_label")


@admin.register(Menu)
class MenuAdmin(admin.ModelAdmin):
    form = BaseMenuForm
    list_display = ('__str__', 'action',)


@admin.register(Permission)
class PermisionAdmin(admin.ModelAdmin):
    model = Permission
    form = BasePermissionForm
    list_display = ("name", "codename", "__str__")

    def __init__(self, model, admin_site):
        self.model._meta.verbose_name = "permiso específico"
        self.model._meta.verbose_name_plural = "permisos específicos"
        super().__init__(model, admin_site)

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        ct = ContentType.objects.get_for_model(self.model)
        codenames = ("add_permission","change_permission","delete_permission","view_permission")
        qs = qs.filter(content_type=ct).exclude(codename__in=codenames)
        return qs

"""



# Python
import inspect
from importlib import import_module


def import_class(module_name, class_name):
    cls = None
    try:
        module = import_module(module_name)
        members = inspect.getmembers(module, inspect.isclass)
        for name, klass in members:
            if name == class_name:
                cls = klass
                break
    except ModuleNotFoundError as error:
        print("Not found %s" % module_name)
    return cls


def import_mixin(name):
    mixin = import_class("superadmin.mixins", name)
    return mixin


def import_all_mixins():
    mixins = list()
    names = (
        "PermissionRequiredMixin",
        "BreadcrumbMixin",
        "UrlMixin",
        "TemplateMixin",
    )
    for name in names:
        mixin = import_mixin(name)
        if mixin:
            mixins.append(mixin)

    return mixins


def get_user_menu(menu_list, user):
    menus = list()
    for menu in menu_list:
        obj_menu = {
            "name": menu.name,
            "url": menu.get_url(),
            "icon": menu.icon_class or "",
            "submenus": get_user_menu(menu.submenus.all(), user),
            "is_root": not menu.parent,
            "is_group": menu.is_group,
        }

        if not obj_menu["submenus"] and (
            menu.is_group or not menu.action.has_permissions(user)
        ):
            continue

        menus.append(obj_menu)
    return menus



import yaml

# Django
from django.core.management.base import BaseCommand, CommandError

# Local
from .base import build_menu


class Command(BaseCommand):
    help = "Create base menu mapping all apps"

    def add_arguments(self, parser):
        parser.add_argument(
            "--build",
            action="store_true",
            help="Build menu",
        )

    def handle(self, *args, **options):
        """
        apps = {}
        for model in site._registry:
            if model._meta.app_config in apps:
                apps[model._meta.app_config].append(model)
            else:
                apps[model._meta.app_config] = [model]
        """
        try:
            with open("menu.yaml") as f:
                data = yaml.load(f, Loader=yaml.FullLoader)
        except FileNotFoundError:
            raise CommandError("Must be a menu.yaml file in the root of the project")
        except yaml.parser.ParserError as e:
            raise CommandError(
                f"""An error has occurred {e.context}:
                - {e.problem}"""
            )
        build_menu(data)
        self.stdout.write(self.style.SUCCESS("Successfully menu was created"))



""" Models for buid menus """

# Django
from django.db import models
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from django.utils.text import slugify
from django.apps import apps
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.conf import settings

# Local
from . import site
from .utils import import_class


class Action(models.Model):
    class ToChoices(models.IntegerChoices):
        MODEL = 1, "Modelo"
        CLASSVIEW = 2, "Vista"

    to = models.PositiveSmallIntegerField(
        choices=ToChoices.choices, verbose_name="acción hacia un"
    )
    app_label = models.CharField(max_length=128, verbose_name="aplicación")
    name = models.CharField(max_length=128, verbose_name="nombre de la acción")
    element = models.CharField(max_length=128, verbose_name="elemento accionado")
    permissions = models.ManyToManyField(
        Permission, blank=True, verbose_name="permisos específicos"
    )

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = "acción"
        verbose_name_plural = "acciones"
        unique_together = ("app_label", "element")
        ordering = ("name",)

    @property
    def model(self):
        if self.to != self.ToChoices.MODEL:
            return None
        try:
            model_class = apps.get_model(self.app_label, self.element)
            return model_class
        except LookupError:
            return None

    @property
    def view(self):
        if self.to != self.ToChoices.CLASSVIEW:
            return None
        try:
            app_config = apps.get_app_config(self.app_label)
            view_class = import_class(f"{app_config.name}.views", self.element)
            return view_class
        except (LookupError):
            return None

    def get_permissions(self):
        return [f"auth.{perm.codename}" for perm in self.permissions.all()]

    def has_permissions(self, user):
        if not user.is_authenticated or not user.is_active:
            return False
        if user.is_superuser:
            pass
            # return True

        if self.to == self.ToChoices.MODEL:
            basic_perms = any(
                user.has_perm(f"{self.app_label}.{perm}_{self.element}")
                for perm in ("view", "add", "change", "delete")
            )
        else:
            basic_perms = True
        specific_perms = all(user.has_perm(perm) for perm in self.get_permissions())

        return basic_perms and specific_perms


class Menu(models.Model):
    """Models for menu"""

    parent = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="submenus",
        on_delete=models.CASCADE,
        verbose_name="menú padre",
    )
    name = models.CharField(max_length=128, verbose_name="nombre")
    route = models.CharField(
        max_length=512, unique=True, editable=False, verbose_name="ruta de acceso"
    )
    action = models.ForeignKey(Action, on_delete=models.CASCADE, verbose_name="acción")
    icon_class = models.CharField(
        max_length=128, blank=True, null=True, verbose_name="clase css del ícono"
    )
    is_group = models.BooleanField(default=False, editable=False, verbose_name="agrupa")
    sequence = models.PositiveSmallIntegerField(verbose_name="secuencia")
    is_active = models.BooleanField(default=True, verbose_name="activo?")

    class Meta:
        ordering = ("sequence",)

    def __str__(self):
        return f"{self.name} | {self.get_route()}"

    def get_route(self):
        route = (
            f"{self.parent.get_route()}/{slugify(self.name)}"
            if self.parent
            else slugify(self.name)
        )
        return route

    def get_url(self):
        url_name = None
        reverse_url = True
        if self.action.to == Action.ToChoices.MODEL:
            model_class = self.action.model
            if model_class and site.is_registered(model_class):
                model_site = site.get_modelsite(model_class)
                url_name = model_site.get_url_name("list")
        else:
            # url_name = f"site:{slugify(self.name)}"
            url_name = f"/{self.get_route()}"
            reverse_url = False
        try:
            if reverse_url:
                url_name = reverse(url_name)
            return url_name
        except NoReverseMatch:
            if settings.DEBUG and url_name:
                print("DEBUG: Not found url for %s" % url_name)
        return "/page/not-found/"



""" """
# Django
from django.views.generic import View
from django.views.generic import DeleteView as BaseDeleteView
from django.http import HttpResponseForbidden, JsonResponse

# Local
from .base import SiteView, get_base_view
from ..shortcuts import get_urls_of_site
from ..utils import import_all_mixins


class DeleteMixin:
    """Definimos la clase que utilizará el modelo"""

    action = "delete"

    def get_slug_field(self):
        return self.site.slug_field or super().get_slug_field()

    def get_success_url(self):
        urls = get_urls_of_site(self.site, object=self.object)
        return urls.get(self.site.delete_success_url)


class DeleteView(SiteView):
    def view(self, request, *args, **kwargs):
        """Crear la List View del modelo"""
        # Class
        mixins = import_all_mixins() + [DeleteMixin]
        View = get_base_view(BaseDeleteView, mixins, self.site)

        # Set attributes
        View.__bases__ = (*self.site.delete_mixins, *View.__bases__)
        view = View.as_view()
        return view(request, *args, **kwargs)


class MassDeleteView(View):
    site = None
    http_method_names = [
        "post",
    ]

    def post(self, request, *args, **kwargs):
        model = self.site.model
        ids = request.POST.getlist("ids")

        if not self.request.user.has_perm(
            f"{model._meta.app_label}.delete_{model._meta.model_name}"
        ):
            return HttpResponseForbidden()

        objects = model.objects.filter(id__in=ids)
        objects.delete()

        return JsonResponse(
            {"success": f"{len(object)} objectos eliminados."}, status=200
        )



""" List view engine"""

# Python
import operator
from functools import reduce

# Django
from django.core.paginator import InvalidPage
from django.db.models import Q
from django.http import Http404
from django.utils.translation import gettext as _
from django.views.generic import ListView as BaseListView


# Local

from .base import SiteView, get_base_view
from ..shortcuts import get_urls_of_site
from ..utils import import_mixin, import_all_mixins

# Utilities
from ..services import FieldService


class ListMixin:
    """Define class"""

    allow_empty = True
    action = "list"

    def get_queryset(self):
        queryset = super().get_queryset()
        search_params = self.request.GET
        if search_params:
            params = search_params.dict()
            search = params.pop("search", None)
            params.pop("page", None)
            params.pop("paginate_by", None)
            model_site = self.site
            if (
                search
                and hasattr(model_site, "search_params")
                and isinstance(model_site.search_params, (list, tuple))
                and model_site.search_params
            ):
                search = search.replace("+", ",").replace(";", ",")
                search_split = search.split(",")
                for search_value in search_split:
                    filters = {
                        key: search_value.strip() for key in model_site.search_params
                    }
                    params.update(**filters)
                    args = [Q(**{key: value}) for key, value in filters.items()]
                    queryset = queryset.filter(reduce(operator.__or__, args))
        return queryset

    def paginate_queryset(self, queryset, page_size):
        """Paginate the queryset, if needed."""
        paginator = self.get_paginator(
            queryset,
            page_size,
            orphans=self.get_paginate_orphans(),
            allow_empty_first_page=self.get_allow_empty(),
        )
        page_kwarg = self.page_kwarg
        page = self.kwargs.get(page_kwarg) or self.request.GET.get(page_kwarg) or 1
        try:
            page_number = int(page)
        except ValueError:
            if page == "last":
                page_number = paginator.num_pages
            else:
                raise Http404(
                    _("Page is not “last”, nor can it be converted to an int.")
                )
        try:
            if page_number > paginator.num_pages:
                page_number = paginator.num_pages
            page = paginator.page(page_number)
            return paginator, page, page.object_list, page.has_other_pages()
        except InvalidPage as e:
            raise Http404(
                _("Invalid page (%(page_number)s): %(message)s")
                % {"page_number": page_number, "message": str(e)}
            )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(self.site.list_extra_context)

        opts = {
            "fields": self.get_list_fields(),
            "rows": self.get_rows(context["object_list"]),
            "page_start_index": context["page_obj"].start_index()
            if context["is_paginated"]
            else 1,
            "page_end_index": context["page_obj"].end_index()
            if context["is_paginated"]
            else context["object_list"].count(),
            "total_records": context["paginator"].count
            if context["is_paginated"]
            else context["object_list"].count(),
        }

        if "site" in context:
            context["site"].update(opts)
        else:
            context.update({"site": opts})

        return context

    def get_paginate_by(self, queryset):
        paginate_by = self.request.GET.get("paginate_by")
        if paginate_by:
            return paginate_by
        return super().get_paginate_by(queryset)

    def get_list_fields(self):
        fields = [
            (name, FieldService.get_field_label(self.model, name))
            for name in self.site.list_fields
        ]
        return fields

    def get_editable_fields(self):
        fields = [
            (name, FieldService.get_field_label(self.model, name))
            for name in self.site.form_class._meta.fields
        ]
        return fields

    def get_rows(self, queryset):
        rows = [
            {
                "instance": instance,
                "values": self.get_values(instance),
                "urls": get_urls_of_site(
                    self.site, object=instance, user=self.request.user
                ),
            }
            for instance in queryset
        ]
        return rows

    def get_values(self, instance):
        values = [
            FieldService.get_field_value(instance, name)
            for name in self.site.list_fields
        ]
        return values


class ListView(SiteView):
    def view(self, request, *args, **kwargs):
        """Crear la List View del modelo"""
        # Class
        FilterMixin = import_mixin("FilterMixin")
        mixins = import_all_mixins() + [FilterMixin, ListMixin]
        View = get_base_view(BaseListView, mixins, self.get_site())

        # Set attriburtes
        View.paginate_by = self.site.paginate_by

        View.__bases__ = (*self.site.list_mixins, *View.__bases__)

        view = View.as_view()
        return view(request, *args, **kwargs)



# Django
from django.views.generic import TemplateView, View
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.contrib import messages

# Utils
from ..utils import get_user_menu


class SiteView(View):
    site = None

    def dispatch(self, request, *args, **kwargs):
        return self.view(request, *args, **kwargs)

    def get_site(self):
        return self.site


class ModuleView(TemplateView):
    """Clase para definir las vistas de los módulos de aplicaciones"""

    menu = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        opts = {"title": self.menu.name}
        if "site" in context:
            context["site"].update(opts)
        else:
            context.update({"site": opts})

        data = {
            "object_list": get_user_menu(self.menu.submenus.all(), self.request.user)
        }

        context.update(data)

        return context

    def get_template_names(self):
        template_name = "superadmin/module_list.html"
        if hasattr(settings, "MODULE_TEMPLATE_NAME"):
            template_name = settings.MODULE_TEMPLATE_NAME
        return [template_name]


def get_base_view(ClassView, mixins, site):
    class View(ClassView):
        def get_context_data(self, **kwargs):
            context = super().get_context_data(**kwargs)
            opts = {
                "title": self.model._meta.verbose_name_plural,
                "app_name": self.model._meta.app_label,
                "model_name": self.model._meta.model_name,
            }

            if "site" in context:
                context["site"].update(opts)
            else:
                context.update({"site": opts})
            return context

        def form_valid(self, form):
            messages.success(self.request, "Se ha guardado correctamente.")
            return super().form_valid(form)

    View.__bases__ = (*mixins, *View.__bases__)
    View.site = site
    View.model = site.model
    View.queryset = site.queryset
    return View



# Shortcuts
from superadmin.shortcuts import get_urls_of_site


class UrlMixin:
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        object = None
        if self.action != "list":
            object = self.object
        urls = get_urls_of_site(self.site, object=object, user=self.request.user)
        if "site" in context:
            context["site"].update({"urls": urls})
        else:
            context.update({"site": {"urls": urls}})
        return context



# Python
from functools import reduce

# Django
from django.forms import BaseModelForm
from django.forms.models import ModelFormMetaclass as DjangoModelFormMetaclass
from django.contrib.admin.utils import flatten
from django.core.exceptions import ImproperlyConfigured


class ModelFormMetaclass(DjangoModelFormMetaclass):
    def __new__(mcs, name, bases, attrs):
        fieldsets = None
        if "Meta" in attrs and hasattr(attrs["Meta"], "fieldsets"):
            fieldsets = attrs["Meta"].fieldsets
            fields = mcs.__fields__(fieldsets)
            if hasattr(attrs["Meta"], "fields"):
                fields = fields + attrs["Meta"].fields
            attrs["Meta"].fields = fields
        new_class = super().__new__(mcs, name, bases, attrs)
        if fieldsets:
            new_class._meta.fieldsets = fieldsets
        return new_class

    def __fields__(fieldsets):
        if isinstance(fieldsets, (list, tuple)):
            fields = flatten(fieldsets)
        elif isinstance(fieldsets, dict):
            fields = reduce(
                lambda acc, fieldset: acc + flatten(fieldset), fieldsets.values(), []
            )
        else:
            raise ImproperlyConfigured(
                "The fieldsets must be an instance of list, tuple or dict"
            )
        return tuple(fields)


class ModelForm(BaseModelForm, metaclass=ModelFormMetaclass):
    def parse(self, fieldset):
        def wrap(fields):
            fields = fields if isinstance(fields, (list, tuple)) else [fields]
            return {
                "bs_cols": int(12 / len(fields)),
                "fields": [self[field] for field in fields],
            }

        fieldset_list = list(map(wrap, fieldset))
        return fieldset_list

    def get_fieldsets(self):
        fieldsets_list = self._meta.fieldsets
        fieldsets = (
            [(None, fieldsets_list)]
            if isinstance(fieldsets_list, (list, tuple))
            else fieldsets_list.items()
        )
        fieldsets = [
            {"title": title or "", "fieldset": self.parse(fieldset)}
            for title, fieldset in fieldsets
        ]

        return fieldsets

    def has_fieldsets(self):
        return hasattr(self._meta, "fieldsets")
